/*
 * To change this license header, choose License Headers in Project Properties.
 * To change this template file, choose Tools | Templates
 * and open the template in the editor.
 */
package gibbs;
import java.util.HashMap;
import java.util.ArrayList;
import java.util.Set;
import java.util.Arrays;
import java.lang.Math;
import java.io.File;
import java.io.FileOutputStream;
import java.io.OutputStreamWriter;
import data.*;
import java.util.Collection;
import java.util.concurrent.ArrayBlockingQueue;
import java.util.concurrent.Callable;
import java.util.concurrent.ThreadPoolExecutor;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.atomic.*;
import utility.Tools;
import utility.ArrayIndexComparator;
/**
 *
 * @author Liangzhe
 */
public class GibbsSampler {
    public double beta,alpha,zeta,eta,xi,sigma,u,tor,rho;
    public int num_vocab,num_users,K,Z,n_itr,burn_in,sam_interval,likelihood_interval;
    public int n_threads;
    public ArrayList<Integer[]> tasks=new ArrayList<Integer[]>();
    public int poolSize;
    public int maxPoolSize;
    public long keepAliveTime=10;
    public ThreadPoolExecutor threadPool=null;
    public ArrayBlockingQueue<Runnable> queue;//=new ArrayBlockingQueue<Runnable>(5);
    public AtomicInteger t_index=new AtomicInteger(0);
    public final AtomicInteger t_iter=new AtomicInteger(0);
    
    public String save_dir;
    public HashMap<Integer,User> users;
    public HashMap<Integer,String> user_map_back;
    public HashMap<Integer,String> word_map_back;
    public Tools tls=new Tools();
    public double[][] I,A,theta,phi;
    public double[] M; //,pi;
    public ArrayList<HashMap<Integer,Integer>> c_u;//member of each community
    public double leave_out=0.05;//the probability mass left out in A for non-members
    //public double perc=0.75;//the percentage of iterations when the discount starts decreasing
    //linearly to zero
    public double discount;
    public int[] count_epsilon_1;
    //public AtomicInteger[] count_epsilon_1;
    //public int[] count_epsilon_0;
    public int[] count_r_1;
    //public AtomicInteger[] count_r_1;
    public AtomicInteger count_s_1;
    public AtomicInteger count_s_0;
    public AtomicInteger[][] count_ee;
    public AtomicInteger[] count_ee_all;
    public AtomicInteger[][] count_c_star;
    public AtomicInteger[] count_c_star_all;
    public AtomicInteger[][] count_c_y;
    public AtomicInteger[] count_c_y_all;
    public AtomicInteger[] count_w_0;//background topic word count
    public AtomicInteger[][] count_w_1;//other topic word count
    public AtomicInteger count_w_all_0;
    public AtomicInteger[] count_w_all_1;
    public AtomicInteger[][] count_z;
    public AtomicInteger[] count_z_all;
    public HashMap<Integer,Double> total_likelihood;
    public HashMap<Integer,Double> text_likelihood;
    public HashMap<Integer,Double> link_likelihood;
    public ArrayList<Integer> likelihood_iters;
    public AtomicInteger count_lambda_1;
    public AtomicInteger count_lambda_0;
    //public AtomicInteger count_celeb_all;
    //public AtomicInteger[] count_celeb;
    public int rid;

    public GibbsSampler(){}
    public void setup(int n_threads,int rid, HashMap<Integer,User> users, HashMap<Integer,String> user_map_back,
            HashMap<Integer,String> word_map_back, int num_vocab, int num_users,
            int K, int Z, int n_itr, int burn_in, int sam_interval, 
            int likelihood_interval, double tor,double discount, String save_dir){
        this.n_threads=n_threads;
        this.poolSize=n_threads;
        this.maxPoolSize=n_threads;
        queue=new ArrayBlockingQueue<Runnable>(n_threads);
        threadPool=new ThreadPoolExecutor(poolSize,maxPoolSize,keepAliveTime,TimeUnit.SECONDS,queue);

        this.rid=rid;
        this.users=users;
        this.user_map_back=user_map_back;
        this.word_map_back=word_map_back;
        this.num_vocab=num_vocab;
        this.num_users=num_users;
        this.K=K;
        this.Z=Z;
        this.n_itr=n_itr;
        this.burn_in=burn_in;
        this.sam_interval=sam_interval;
        this.likelihood_interval=likelihood_interval;
        this.tor=tor;
        this.discount=discount;
        this.save_dir=save_dir;
        int span=(int)Math.round(num_users/n_threads);
        for (int i=0;i<n_threads-1;i++){
            tasks.add(new Integer[]{i*span,(i+1)*span});
        }
        tasks.add(new Integer[]{(n_threads-1)*span,num_users});
    }
    public void run() throws InterruptedException{
        printParam();
        System.out.println("Initialization...");
        initialize();
        System.out.println("Start sampling...");
        Collection<Callable<Void>> callables=new ArrayList<Callable<Void>>();
        for (int iter=0;iter<n_itr;iter++){
            long startTime=System.currentTimeMillis();
            System.out.println("run "+rid+" iteration "+iter);
            callables.clear();
            t_index.set(0);
            for(int i=0;i<n_threads;i++){
                callables.add(new Callable<Void>(){
                    public Void call(){
                        int local=t_index.incrementAndGet();
                        //System.out.println(t_iter.get()+" "+tasks.get((local-1))[0]+" "+tasks.get((local-1))[1]);
                        nextSampling(t_iter.get(),tasks.get(local-1));
                        return null;
                    }
                });
            }
            threadPool.invokeAll(callables);
            t_iter.incrementAndGet();
            //update the c_u community membership
            for(int i=0;i<c_u.size();i++){
                c_u.get(i).clear();
            }
            for(int i=0;i<num_users;i++){
                int com=users.get(i).c;
                c_u.get(com).put(i,1);
            }
            //sample M
            sample_M();
            //sample u
            //sample_u();
            //checkcount();
            long endTime=System.currentTimeMillis();
            System.out.println("Iteration Running time: "+(endTime-startTime)/1000+" secs.");
            if(iter>=burn_in && iter%sam_interval==0){
                sample_others(iter);
                update_likelihood(iter);
                save_results(iter);
            }
        }
        threadPool.shutdown();
        System.out.println("Gibbs sampling done.");
    }
    
    public void checkcount(){
        for (int i=0;i<count_ee.length;i++){
            for(int j=0;j<count_ee[i].length;j++){
                if (count_ee[i][j].get()<0){
                    System.out.println("count_ee "+i+","+j+" "+count_ee[i][j]);
                    System.exit(1);
                }
            }
        }
        for (int i=0;i<count_c_star.length;i++){
            for(int j=0;j<count_c_star[i].length;j++){
                if (count_c_star[i][j].get()<0){
                    System.out.println("count_c_star "+i+","+j+" "+count_c_star[i][j]);
                    System.exit(1);
                }
            }
        }
        for (int i=0;i<count_c_y.length;i++){
            for(int j=0;j<count_c_y[i].length;j++){
                if (count_c_y[i][j].get()<0){
                    System.out.println("count_c_y "+i+","+j+" "+count_c_y[i][j]);
                    System.exit(1);
                }
            }
        }
        for (int i=0;i<count_w_1.length;i++){
            for(int j=0;j<count_w_1[i].length;j++){
                if (count_w_1[i][j].get()<0){
                    System.out.println("count_w_1 "+i+","+j+" "+count_w_1[i][j]);
                    System.exit(1);
                }
            }
        }
        for (int i=0;i<count_z.length;i++){
            for(int j=0;j<count_z[i].length;j++){
                if (count_z[i][j].get()<0){
                    System.out.println("count_z "+i+","+j+" "+count_z[i][j]);
                    System.exit(1);
                }
            }
        }
        for (int i=0;i<count_epsilon_1.length;i++){
            //if (count_epsilon_1[i].get()<0){
            if (count_epsilon_1[i]<0){
                System.out.println("count_epsilon_1 "+i+" "+count_epsilon_1[i]);
                System.exit(1);
            }
        }
        for (int i=0;i<count_r_1.length;i++){
            //if (count_r_1[i].get()<0){
            if(count_r_1[i]<0){
                System.out.println("count_r_1 "+i+" "+count_r_1[i]);
                System.exit(1);
            }
        }
        for (int i=0;i<count_ee_all.length;i++){
            if (count_ee_all[i].get()<0){
                System.out.println("count_ee_all "+i+" "+count_ee_all[i]);
                System.exit(1);
            }
        }
        for (int i=0;i<count_c_star_all.length;i++){
            if (count_c_star_all[i].get()<0){
                System.out.println("count_c_star_all "+i+" "+count_c_star_all[i]);
                System.exit(1);
            }
        }
        for (int i=0;i<count_c_y_all.length;i++){
            if (count_c_y_all[i].get()<0){
                System.out.println("count_c_y_all "+i+" "+count_c_y_all[i]);
                System.exit(1);
            }
        }
        for (int i=0;i<count_w_0.length;i++){
            if (count_w_0[i].get()<0){
                System.out.println("count_w_0 "+i+" "+count_w_0[i]);
                System.exit(1);
            }
        }
        for (int i=0;i<count_w_all_1.length;i++){
            if (count_w_all_1[i].get()<0){
                System.out.println("count_w_all_1 "+i+" "+count_w_all_1[i]);
                System.exit(1);
            }
        }
        for (int i=0;i<count_z_all.length;i++){
            if (count_z_all[i].get()<0){
                System.out.println("count_z_all "+i+" "+count_z_all[i]);
                System.exit(1);
            }
        }
    }
    public void update_likelihood(int iter){
        double w_like=0.0;
        double l_like=0.0;
        User curUser;
        Document curDocument;
        Link curLink;
        Word curWord;
        int c,z,s,wid,e,epsilon,c_star, lambda;
        for (int i=0;i<num_users;i++){
            curUser=users.get(i);
            c=curUser.c;
            for(int j=0;j<curUser.no_docs;j++){
                curDocument=curUser.docs.get(j);
                z=curDocument.z;
                for (int t=0;t<curDocument.no_words;t++){
                    curWord=curDocument.words.get(t);
                    wid=curWord.wid;
                    s=curWord.s;
                    if(s==0){
                        w_like+=Math.log(phi[Z][wid]);
                    }else{
                        w_like+=Math.log(phi[z][wid]);
                    }
                }
            }
            for(int j=0;j<curUser.no_links;j++){
                curLink=curUser.links.get(j);
                e=curLink.target;
                epsilon=curLink.epsilon;
                lambda = curLink.lambda;
                //System.out.println(epsilon + ", " + lambda);
                c_star=curLink.c_star;
                if(epsilon==0){
                    l_like+=Math.log(A[c_star][e]);
                    //System.out.println(A[c_star][e]);
                }else if (lambda == 0){
                    l_like+=Math.log(A[c][e]);
                    //System.out.println(A[c][e]);
                }else{
                    l_like += Math.log(1.0 / num_users);
                    //System.out.println(1.0 / num_users);
                }
            }
        }
        likelihood_iters.add(iter);
        text_likelihood.put(iter,-w_like);
        link_likelihood.put(iter, -l_like);
        total_likelihood.put(iter, -w_like-l_like);
    }
    public void save_results(int iter){
        System.out.println("saving results for this iteration...");
        try{
            String folder=save_dir+iter;
            File file=new File(folder);
            if(!file.exists()){
                file.mkdirs();
            }
            String filename=save_dir+iter+"/likelihood.txt";
            OutputStreamWriter out_f=new OutputStreamWriter(
            new FileOutputStream(new File(filename)));
            out_f.write("iteration\tw+e\tw\te\n");
            for(int i=0;i<likelihood_iters.size();i++){
                int key=likelihood_iters.get(i);
                out_f.write(key+"\t"+total_likelihood.get(key)+"\t"+
                        text_likelihood.get(key)+"\t"+link_likelihood.get(key)+"\n");
            }
            out_f.flush();
            out_f.close();
            
            filename=save_dir+iter+"/members.txt";
            out_f=new OutputStreamWriter(
            new FileOutputStream(new File(filename)));
            for(int i=0;i<c_u.size();i++){
                out_f.write(i+"\n");
                for(int uid:c_u.get(i).keySet()){
                    out_f.write(user_map_back.get(uid)+":v="+users.get(uid).v+"\n");
                }
                out_f.write("\n");
            }
            out_f.flush();
            out_f.close();
            
            filename=save_dir+iter+"/phi.txt";
            out_f=new OutputStreamWriter(
            new FileOutputStream(new File(filename)));
            for(int i=0;i<Z+1;i++){
                out_f.write("topic "+i+"\n");
                for(int j=0;j<num_vocab-1;j++){
                    out_f.write(phi[i][j]+"\t");
                }
                out_f.write(phi[i][num_vocab-1]+"");
                out_f.write("\n");
            }
            out_f.flush();
            out_f.close();
            
            filename=save_dir+iter+"/A.txt";
            out_f=new OutputStreamWriter(
            new FileOutputStream(new File(filename)));
            for(int i=0;i<K;i++){
                out_f.write("community "+i+"\n");
                for(int j=0;j<num_users-1;j++){
                    out_f.write(A[i][j]+"\t");
                }
                out_f.write(A[i][num_users-1]+"");
                out_f.write("\n");
            }
            out_f.flush();
            out_f.close();
            
            ArrayIndexComparator aic=new ArrayIndexComparator();
            Integer[] indexes;
            filename=save_dir+iter+"/freq_words.txt";
            out_f=new OutputStreamWriter(
            new FileOutputStream(new File(filename)));
            for(int i=0;i<Z+1;i++){
                out_f.write("topic "+i+"\n");
                aic.setArray(phi[i]);
                indexes=aic.getIndex();
                Arrays.sort(indexes,aic);
                for (int j=0;j<15;j++){
                    out_f.write(word_map_back.get(indexes[j])+"\t"+phi[i][indexes[j]]+"\n");
                }
                out_f.write("\n");
            }
            out_f.flush();
            out_f.close();
            
            filename=save_dir+iter+"/fam_users.txt";
            out_f=new OutputStreamWriter(
            new FileOutputStream(new File(filename)));
            for(int i=0;i<K;i++){
                out_f.write("community "+i+"\n");
                aic.setArray(A[i]);
                indexes=aic.getIndex();
                Arrays.sort(indexes,aic);
                for (int j=0;j<15;j++){
                    out_f.write(user_map_back.get(indexes[j])+"\t"+A[i][indexes[j]]+"\n");
                }
                out_f.write("\n");
            }
            out_f.flush();
            out_f.close();
            
            
            filename=save_dir+iter+"/probs.txt";
            out_f=new OutputStreamWriter(
            new FileOutputStream(new File(filename)));
            out_f.write("u="+u+"\n");
	    out_f.write("rho=" + rho + "\n");
            out_f.write("M=[");
            for(double i:M){
                out_f.write(" "+i);
            }
            out_f.write("]\n");
            out_f.write("theta=[");
            for(int i=0;i<K;i++){
                out_f.write("[");
                for(int j=0;j<Z;j++){
                    out_f.write(" "+theta[i][j]);
                }
                out_f.write("]");
                if (i!=K-1){
                    out_f.write("\n");
                }
            }
            out_f.write("]\n");
            out_f.write("I=[");
            for(int i=0;i<K;i++){
                out_f.write("[");
                for (int j=0;j<K;j++){
                    out_f.write(" "+I[i][j]);
                }
                out_f.write("]");
                if (i!=K-1){
                    out_f.write("\n");
                }
            }
            out_f.flush();
            out_f.close();
        }catch(Exception e){
            e.printStackTrace();
        }
    }
    public void sample_others(int iter){
        //sample phi,A,I,theta
        sample_phi();
        sample_A(iter);
        sample_I();
        sample_theta();
    }
    public void sample_theta(){
        double pp1,pp1_all,p1;
        //double[] p=new double[Z];
        for (int i=0;i<K;i++){
            for (int j=0;j<Z;j++){
                pp1=count_z[j][i].get();
                pp1_all=count_z_all[i].get();
                p1=(alpha+pp1)/(alpha*Z+pp1_all);
                //p[j]=p1;
                theta[i][j]=p1;
            }
            //p=tls.normalize(p);
            //theta[i]=p;
        }
        theta=tls.normalize(theta);
    }
    public void sample_I(){
        double pp1,pp1_all,p1;
        //double[] p=new double[K];
        for(int i=0;i<K;i++){
            for(int j=0;j<K;j++){
                pp1=count_c_y[j][i].get()+count_c_star[j][i].get();
                pp1_all=count_c_y_all[i].get()+count_c_star_all[i].get();
                p1=(zeta+pp1)/(zeta*K+pp1_all);
                //p[j]=p1;
                I[i][j]=p1;
            }
            //p=tls.normalize(p);
            //I[i]=p;
        }
        I=tls.normalize(I);
    }
    public void sample_A(int iter){
        double new_discount=get_current_discount(discount,iter,n_itr);
        double pp1,pp1_all,p1;
        //double[] p=new double[num_users];
        for(int i=0;i<K;i++){
            for(int j=0;j<num_users;j++){
                pp1=count_ee[j][i].get();
                pp1_all=count_ee_all[i].get();
                if(users.get(j).c==i){
                    p1=(eta+pp1)/(eta*num_users+pp1_all);
                }else{
                    p1=(eta+new_discount*pp1)/(eta*num_users+pp1_all);
                }
                //p[j]=p1;
                A[i][j]=p1;
            }
            //p=tls.normalize(p);
            //A[i]=p;
        }
        A=tls.normalize(A);
    }
    public void sample_phi(){
        double pp1,pp1_all,p1;
        //double[] p=new double[num_vocab];
        for(int i=0;i<Z;i++){
            pp1_all=count_w_all_1[i].get();
            for (int j=0;j<num_vocab;j++){
                pp1=count_w_1[j][i].get();
                p1=(beta+pp1)/(beta*num_vocab+pp1_all);
                //p[j]=p1;
                phi[i][j]=p1;
            }
            //p=tls.normalize(p);
            //phi[i]=p;
        }
        //phi=tls.normalize(phi);
        for(int i=0;i<num_vocab;i++){
            pp1=count_w_0[i].get();
            pp1_all=count_w_all_0.get();
            p1=(beta+pp1)/(beta*num_vocab+pp1_all);
            //p[i]=p1;
            phi[Z][i]=p1;
        }
        //p=tls.normalize(p);
        //phi[Z]=p;
        phi=tls.normalize(phi);
    }
    public void nextSampling(int iter,Integer[] interval){
        int s_index=interval[0];
        int e_index=interval[1];
        User curUser;
        //long startTime=System.currentTimeMillis();
        Document curDocument;
        //sample z
        for(int i=s_index;i<e_index;i++){
            curUser=users.get(i);
            for(int j=0;j<curUser.no_docs;j++){
                sample_z(i,j);
            }
        }
        //sample r
        for(int i=s_index;i<e_index;i++){
            curUser=users.get(i);
            for(int j=0;j<curUser.no_docs;j++){
                sample_r(i,j);
            }
        }
        //sample s
        for(int i=s_index;i<e_index;i++){
            curUser=users.get(i);
            for(int j=0;j<curUser.no_docs;j++){
                sample_s(i,j);
            }
        }
        //sample c_y
        for(int i=s_index;i<e_index;i++){
            curUser=users.get(i);
            for(int j=0;j<curUser.no_docs;j++){
                sample_c_y(i,j);
            }
        }
        //sample c_star
        for(int i=s_index;i<e_index;i++){
            curUser=users.get(i);
            for(int j=0;j<curUser.no_links;j++){
                sample_c_star(i,j,iter);
            }
        }
        //sample c
        for(int i=s_index;i<e_index;i++){
            sample_c(i,iter);
        }
        //sample epsilon
        for(int i=s_index;i<e_index;i++){
            curUser=users.get(i);
            for(int j=0;j<curUser.no_links;j++){
                sample_epsilon(i,j,iter);
            }
        }
        //sample lambda
        for (int i=s_index;i<e_index;i++){
            curUser=users.get(i);
            for (int j=0;j<curUser.no_links;j++){
                sample_lambda(i,j,iter);
            }
        }
        //sample v
        for (int i=s_index;i<e_index;i++){
            sample_v(i);
        }
        //sample rho
        sample_rho();
        //long endTime=System.currentTimeMillis();
        //System.out.println((endTime-startTime)/1000);
        
    }
    public void sample_rho(){
        double new_rho = tls.sample_beta(0.5 + count_lambda_0.get(), 0.5 + count_lambda_1.get());
        rho = new_rho;
    }
    public void sample_M(){
        double pp1;
        double[] p=new double[K];
        int pp1_all=num_users;
        for(int i=0;i<K;i++){
            pp1=c_u.get(i).size();
            p[i]=(sigma+pp1)/(K*sigma+pp1_all);
        }
        M=tls.normalize(p);
    }
    public void sample_lambda(int uid,int lid,int iter){
        double new_discount=get_current_discount(discount,iter,n_itr);
        User curUser=users.get(uid);
        int e,c_star,epsilon,lambda;
        int c=curUser.c;
        Link curLink=curUser.links.get(lid);
        e=curLink.target;
        c_star=curLink.c_star;
        epsilon=curLink.epsilon;
        lambda=curLink.lambda;
        double p1,pp1,pp1_all,p2,pp2,pp2_all;
        double[] p=new double[2];
        //if (epsilon==0){
        //    p[0]=rho;
        //    p[1]=1.0-rho;
        //}else{
        pp1=count_ee[e][c].get();
        pp1_all=count_ee_all[c].get();
        if (lambda==0){
            pp1-=1;
            pp1_all-=1;
        }
        if (users.get(e).c==c){
            p1=(eta+pp1)/(num_users*eta+pp1_all);
        }else{
            p1=(eta+new_discount*pp1)/(num_users*eta+pp1_all);
        }
        //pp2=count_celeb[e].get();
        //pp2_all=count_celeb_all.get();
        //if (lambda==1){
        //    pp2-=1;
        //    pp2_all-=1;
        //}
        //p2=(eta+pp2)/(num_users*eta+pp2_all);
        p2 = 1.0 / num_users;//random following
        p[0]=rho*p1;
        p[1]=(1-rho)*p2;
        //}
        int new_lambda=tls.sample_multi(p);
        if (new_lambda!=lambda){
            curLink.lambda=new_lambda;
            if (epsilon==1){
                if (lambda==0){
                    //originally own com, now random
                    count_ee[e][c].decrementAndGet();
                    count_ee_all[c].decrementAndGet();
                    count_lambda_0.decrementAndGet();
                    count_lambda_1.incrementAndGet();
                }else if(lambda==1){
                    //originally random, now from own com
                    count_ee[e][c].incrementAndGet();
                    count_ee_all[c].incrementAndGet();
                    count_lambda_0.incrementAndGet();
                    count_lambda_1.decrementAndGet();
                }
            }
            //if (lambda==0){
            //    count_lambda_0.decrementAndGet();
            //    count_lambda_1.incrementAndGet();
            //}else{
            //    count_lambda_0.incrementAndGet();
            //    count_lambda_1.decrementAndGet();
            //}
        }
    }
    public void sample_epsilon(int uid,int lid,int iter){
        double new_discount=get_current_discount(discount,iter,n_itr);
        User curUser=users.get(uid);
        int c=curUser.c;
        int e,c_star,epsilon, lambda;
        double p1,pp1,pp1_all,pp2,pp2_all,p2;
        double[] p=new double[2];
        Link curLink=curUser.links.get(lid);
        e=curLink.target;
        c_star=curLink.c_star;
        epsilon=curLink.epsilon;
        lambda = curLink.lambda;
        pp1=count_ee[e][c_star].get();
        pp1_all=count_ee_all[c_star].get();
        if (epsilon==0){
            pp1--;
            pp1_all--;
        }
        if(users.get(e).c==c_star){//e is a member of c_star
            p1=(eta+pp1)/(num_users*eta+pp1_all);
        }else{//e is not a member of c_star
            p1=(eta+new_discount*pp1)/(eta*num_users+pp1_all);
        }
        if (lambda == 0){
            pp2=count_ee[e][c].get();
            pp2_all=count_ee_all[c].get();
            if (epsilon==1){
                pp2--;
                pp2_all--;
            }
            if(users.get(e).c==c){//e is a member of c
                p2=(eta+pp2)/(num_users*eta+pp2_all);
            }else{
                p2=(eta+new_discount*pp2)/(num_users*eta+pp2_all);
            }
        }else{
            p2 = 1.0 / num_users;//following users randomly
        }
        p[0]=(1-curUser.v)*p1;
        p[1]=curUser.v*p2;
        int new_epsilon=tls.sample_multi(p);
        if(new_epsilon!=epsilon){
            curLink.epsilon=new_epsilon;
            if (epsilon==1 && lambda == 0){
                //from own community to influencing community
                count_c_star[c_star][c].incrementAndGet();
                count_c_star_all[c].incrementAndGet();
                count_ee[e][c].decrementAndGet();
                count_ee[e][c_star].incrementAndGet();
                count_ee_all[c].decrementAndGet();
                count_ee_all[c_star].incrementAndGet();
                //count_epsilon_1[uid].decrementAndGet();
                count_epsilon_1[uid]-=1;
                count_lambda_0.decrementAndGet();
            }else if (epsilon == 1 && lambda == 1){
                //from random to influencing community
                count_c_star[c_star][c].incrementAndGet();
                count_c_star_all[c].incrementAndGet();
                count_ee[e][c_star].incrementAndGet();
                count_ee_all[c_star].incrementAndGet();
                //count_epsilon_1[uid].decrementAndGet();
                count_epsilon_1[uid]-=1;
                count_lambda_1.decrementAndGet();
            }else if (epsilon == 0 && lambda == 0){
                //from influencing community to own community
                count_c_star[c_star][c].decrementAndGet();
                count_c_star_all[c].decrementAndGet();
                count_ee[e][c].incrementAndGet();
                count_ee[e][c_star].decrementAndGet();
                count_ee_all[c].incrementAndGet();
                count_ee_all[c_star].decrementAndGet();
                //count_epsilon_1[uid].incrementAndGet();
                count_epsilon_1[uid]+=1;
                count_lambda_0.incrementAndGet();
            }else if (epsilon == 0 && lambda == 1){
                //from influencing community to random 
                count_c_star[c_star][c].decrementAndGet();
                count_c_star_all[c].decrementAndGet();
                count_ee[e][c_star].decrementAndGet();
                count_ee_all[c_star].decrementAndGet();
                //count_epsilon_1[uid].incrementAndGet();
                count_epsilon_1[uid]+=1;
                count_lambda_1.incrementAndGet();
            }
        }
    }
    public void sample_c_star(int uid,int lid,int iter){
        double new_discount=get_current_discount(discount,iter,n_itr);
        User curUser=users.get(uid);
        int c=curUser.c;
        int e,c_star,epsilon, lambda;
        double p1,pp1,pp1_all,pp2,pp2_all,p2;
        double[] p=new double[K];
        Link curLink=curUser.links.get(lid);
        e=curLink.target;
        c_star=curLink.c_star;
        epsilon=curLink.epsilon;
        lambda = curLink.lambda;
        for(int poten_c_star=0;poten_c_star<K;poten_c_star++){
            p1=1.0;//remain unchanged if epsilon=1
            if (epsilon==0){//influence from another community
                pp1=count_ee[e][poten_c_star].get();
                pp1_all=count_ee_all[poten_c_star].get();
                if(poten_c_star==c_star){
                    pp1--;
                    pp1_all--;
                }
                if (users.get(e).c==poten_c_star){//e is a member of poten_c_star
                    p1=(eta+pp1)/(num_users*eta+pp1_all);
                }else{
                    p1=(eta+new_discount*pp1)/(num_users*eta+pp1_all);
                }
            }
            pp2=count_c_star[poten_c_star][c].get()+count_c_y[poten_c_star][c].get();
            pp2_all=count_c_star_all[c].get()+count_c_y_all[c].get();
            if (poten_c_star==c_star && epsilon==0){
                pp2--;
                pp2_all--;
            }
            p2=(zeta+pp2)/(K*zeta+pp2_all);
            p[poten_c_star]=p1*p2;
        }
        int new_c_star=tls.sample_multi(p);
        if (new_c_star!=c_star){
            curLink.c_star=new_c_star;
            if (epsilon==0){
                count_c_star[c_star][c].decrementAndGet();
                count_c_star[new_c_star][c].incrementAndGet();
                count_ee[e][c_star].decrementAndGet();
                count_ee[e][new_c_star].incrementAndGet();
                count_ee_all[c_star].decrementAndGet();
                count_ee_all[new_c_star].incrementAndGet();
            }
        }
    }
    public void sample_v(int uid){
        User curUser=users.get(uid);
        int ds=curUser.no_docs;
        int ls=curUser.no_links;
        //double new_v=tls.sample_beta(0.5+ds-count_r_1[uid].get()+count_epsilon_1[uid].get(),
        //        0.5+count_r_1[uid].get()+ls-count_epsilon_1[uid].get());
        double new_v=tls.sample_beta(0.5+ds-count_r_1[uid]+count_epsilon_1[uid],
                0.5+count_r_1[uid]+ls-count_epsilon_1[uid]);
        curUser.v=new_v;
    }
    public void sample_u(){
        double new_u=tls.sample_beta(0.5+count_s_0.get(),0.5+count_s_1.get());
        u=new_u;
    }
    public void sample_c(int uid,int iter){
        User curUser=users.get(uid);
        Document curDocument;
        Link curLink;
        double p1,p2,p3,p4,pp1,pp1_all,pp2,pp2_all,pp3,pp3_all,pp4,pp4_all;
        double[] p=new double[K];
        int r,z,c_y,c_star,epsilon,e, lambda;
        double new_discount=get_current_discount(discount,iter,n_itr);
        int c=curUser.c;
        for (int poten_c=0;poten_c<K;poten_c++){
            p1=0;
            p2=0;
            p3=0;
            p4=0;
            for (int i=0;i<curUser.no_docs;i++){
                curDocument=curUser.docs.get(i);
                r=curDocument.r;
                z=curDocument.z;
                c_y=curDocument.c_y;
                if(r==0){
                    pp1=count_z[z][poten_c].get();
                    pp1_all=count_z_all[poten_c].get();
                    if (poten_c==c){
                        pp1--;
                        pp1_all--;
                    }
                    p1+=Math.log((alpha+pp1)/(Z*alpha+pp1_all));
                }else{
                    pp3=count_c_y[c_y][poten_c].get();
                    pp3_all=count_c_y_all[poten_c].get();
                    if (poten_c==c){
                        pp3--;
                        pp3_all--;
                    }
                    p3+=Math.log((zeta+pp3)/(K*zeta+pp3_all));
                }
            }
            for (int i=0;i<curUser.no_links;i++){
                curLink=curUser.links.get(i);
                e=curLink.target;
                c_star=curLink.c_star;
                epsilon=curLink.epsilon;
                lambda = curLink.lambda;
                
                if(epsilon == 1 && lambda == 0){//not influence, selected from own community
                    pp2=count_ee[e][poten_c].get();
                    pp2_all=count_ee_all[poten_c].get();
                    if (poten_c==c){
                        pp2--;
                        pp2_all--;
                    }
                    if (users.get(e).c==poten_c){//the followee is from the same community
                        p2+=Math.log((eta+pp2)/(num_users*eta+pp2_all));
                    }else{//from a different community
                        p2+=Math.log((eta+new_discount*pp2)/(num_users*eta+pp2_all));
                    }
                }else if (epsilon == 0){//influenced by another community
                    pp4=count_c_star[c_star][poten_c].get();
                    pp4_all=count_c_star_all[poten_c].get();
                    if (poten_c==c){
                        pp4--;
                        pp4_all--;
                    }
                    p4+=Math.log((zeta+pp4)/(K*zeta+pp4_all));
                }
            }
            //p[poten_c]=Math.pow(Math.E, (p1+p2+p3+p4+Math.log(M[poten_c])));
            p[poten_c] = p1 + p2 + p3 + p4 + Math.log(M[poten_c]);
        }
        //int new_c=tls.sample_multi(p);
        int new_c = tls.sample_multi_log(p);
        if(new_c!=c){
            curUser.c=new_c;
            //c_u.get(c).remove(uid);
            //c_u.get(new_c).put(uid, 1);
            for(int i=0;i<curUser.no_docs;i++){
                curDocument=curUser.docs.get(i);
                r=curDocument.r;
                z=curDocument.z;
                c_y=curDocument.c_y;
                if(r==0){
                    count_z[z][c].decrementAndGet();
                    count_z[z][new_c].incrementAndGet();
                    count_z_all[c].decrementAndGet();
                    count_z_all[new_c].incrementAndGet();
                }else{
                    count_c_y[c_y][c].decrementAndGet();
                    count_c_y[c_y][new_c].incrementAndGet();
                    count_c_y_all[c].decrementAndGet();
                    count_c_y_all[new_c].incrementAndGet();
                }
            }
            for (int i=0;i<curUser.no_links;i++){
                curLink=curUser.links.get(i);
                e=curLink.target;
                c_star=curLink.c_star;
                epsilon=curLink.epsilon;
                lambda = curLink.lambda;
                if (epsilon==0){
                    count_c_star[c_star][c].decrementAndGet();
                    count_c_star[c_star][new_c].incrementAndGet();
                    count_c_star_all[c].decrementAndGet();
                    count_c_star_all[new_c].incrementAndGet();
                }else if (epsilon == 1 && lambda == 0){
                    count_ee[e][c].decrementAndGet();
                    count_ee[e][new_c].incrementAndGet();
                    count_ee_all[c].decrementAndGet();
                    count_ee_all[new_c].incrementAndGet();
                }
            }
        }
    }
    public double get_current_discount_prev(double d, int iter, int n_itr){
        double perc=0.75;
        if(iter<perc*n_itr){
            return d;
        }else{
            double a=-d/(n_itr*(1-perc));
            double b=d/(1-perc);
            return a*iter+b;
        }
    }
    public double get_current_discount(double d, int iter, int n_itr){
        double perc=0.45;
        double perc2 = 0.55;
        if(iter<perc*n_itr){
            return d;
        }else if (iter < perc2 * n_itr){
            double a=-d/(n_itr*(perc2-perc));
            double b=d * perc2 /(perc2 - perc);
            return a*iter+b;
        }else{
            return 0;
        }
    }
    public void sample_c_y(int uid,int did){
        User curUser=users.get(uid);
        Document curDocument=curUser.docs.get(did);
        Word curWord;
        int r=curDocument.r;
        int c=curUser.c;
        int c_y=curDocument.c_y;
        int z=curDocument.z;
        double[] p=new double[K];
        double p1,p2,pp1,pp1_all,pp2,pp2_all;
        for(int poten_c_y=0;poten_c_y<K;poten_c_y++){
            if (r==0){
                pp1=count_z[z][c].get()-1;
                pp1_all=count_z_all[c].get()-1;
            }else{
                pp1=count_z[z][poten_c_y].get();
                pp1_all=count_z_all[poten_c_y].get();
                if (poten_c_y==c_y){
                    pp1--;
                    pp1_all--;
                }
            }
            p1=(alpha+pp1)/(Z*alpha+pp1_all);
            pp2=count_c_y[poten_c_y][c].get()+count_c_star[poten_c_y][c].get();
            if (poten_c_y==c_y && r==1){
                pp2--;
            }
            pp2_all=count_c_y_all[c].get()+count_c_star_all[c].get();
            if (r==1)pp2_all--;
            p2=(zeta+pp2)/(K*zeta+pp2_all);
            p[poten_c_y]=p1*p2;
        }
        int new_c_y=tls.sample_multi(p);
        if (new_c_y!=c_y){
            curDocument.c_y=new_c_y;
            if(r==1){
                count_c_y[c_y][c].decrementAndGet();
                count_c_y[new_c_y][c].incrementAndGet();
                count_z[z][c_y].decrementAndGet();
                count_z[z][new_c_y].incrementAndGet();
                count_z_all[c_y].decrementAndGet();
                count_z_all[new_c_y].incrementAndGet();
            }
        }
    }
    public void sample_s(int uid,int did){
        User curUser=users.get(uid);
        Document curDocument=curUser.docs.get(did);
        Word curWord;
        int r=curDocument.r;
        int c=curUser.c;
        int c_y=curDocument.c_y;
        int z=curDocument.z;
        double pp1,pp1_all,p1;
        for(int i=0;i<curDocument.no_words;i++){
            curWord=curDocument.words.get(i);
            double[] p=new double[2];
            int s=curWord.s;
            int wid=curWord.wid;
            pp1=count_w_0[wid].get();
            pp1_all=count_w_all_0.get();
            if (s==0){
                pp1--;
                pp1_all--;
            }
            p1=u*(beta+pp1)/(num_vocab*beta+pp1_all);
            p[0]=p1;
            pp1=count_w_1[wid][z].get();
            pp1_all=count_w_all_1[z].get();
            if (s==1){
                pp1--;
                pp1_all--;
            }
            p1=(1-u)*(beta+pp1)/(num_vocab*beta+pp1_all);
            p[1]=p1;
            int new_s=tls.sample_multi(p);
            if (new_s!=s){
                curWord.s=new_s;
                if(s==0){
                    count_w_0[wid].decrementAndGet();
                    count_w_1[wid][z].incrementAndGet();
                    count_w_all_0.decrementAndGet();
                    count_w_all_1[z].incrementAndGet();
                    count_s_0.decrementAndGet();
                    count_s_1.incrementAndGet();
                    curDocument.count_s_1++;
                    if (curDocument.word_count.containsKey(wid)){
                        curDocument.word_count.put(wid,curDocument.word_count.get(wid)+1);
                    }else{
                        curDocument.word_count.put(wid,1);
                    }
                }else{
                    count_w_0[wid].incrementAndGet();
                    count_w_1[wid][z].decrementAndGet();
                    count_w_all_0.incrementAndGet();
                    count_w_all_1[z].decrementAndGet();
                    count_s_0.incrementAndGet();
                    count_s_1.decrementAndGet();
                    curDocument.count_s_1--;
                    curDocument.word_count.put(wid,curDocument.word_count.get(wid)-1);
                }
            }
        }
    }
    public void sample_r(int uid,int did){
        User curUser=users.get(uid);
        Document curDocument=curUser.docs.get(did);
        int r=curDocument.r;
        int c=curUser.c;
        int c_y=curDocument.c_y;
        int z=curDocument.z;
        double[] p=new double[2];
        double p1,pp1,pp1_all;
        pp1=count_z[z][c].get();
        pp1_all=count_z_all[c].get();
        if(r==0){
            pp1--;
            pp1_all--;
        }
        if (curDocument.l==0){
            p1=(1-tor+tor*curUser.v)*(alpha+pp1)/(Z*alpha+pp1_all);
        }else{
            p1=curUser.v*tor*(alpha+pp1)/(Z*alpha+pp1_all);
        }
        p[0]=p1;
        pp1=count_z[z][c_y].get();
        pp1_all=count_z_all[c_y].get();
        if(r==0){
            pp1--;
            pp1_all--;
        }
        if (curDocument.l==0){
            p1=tor*(1-curUser.v)*(alpha+pp1)/(Z*alpha+pp1_all);
        }else{
            p1=(1-curUser.v*tor)*(alpha+pp1)/(Z*alpha+pp1_all);
        }
        p[1]=p1;
        int new_r=tls.sample_multi(p);
        if (new_r!=r){
            curDocument.r=new_r;
            if (r==1){
                count_c_y[c_y][c].decrementAndGet();
                count_c_y_all[c].decrementAndGet();
                count_z[z][c_y].decrementAndGet();
                count_z[z][c].incrementAndGet();
                count_z_all[c_y].decrementAndGet();
                count_z_all[c].incrementAndGet();
                //count_r_1[uid].decrementAndGet();
                count_r_1[uid]-=1;
            }else{
                count_c_y[c_y][c].incrementAndGet();
                count_c_y_all[c].incrementAndGet();
                count_z[z][c_y].incrementAndGet();
                count_z[z][c].decrementAndGet();
                count_z_all[c_y].incrementAndGet();
                count_z_all[c].decrementAndGet();
                //count_r_1[uid].incrementAndGet();
                count_r_1[uid]+=1;
            }
        }
    }
    public void sample_z(int uid,int did){
        User curUser=users.get(uid);
        Document curDocument=curUser.docs.get(did);
        Word curWord;
        int r=curDocument.r;
        int c=curUser.c;
        int c_y=curDocument.c_y;
        int z=curDocument.z;
        double[] p=new double[Z];
        for(int poten_z=0;poten_z<Z;poten_z++){
            double p1=0.0,p2;
            double pp1,pp1_all,pp2,pp2_all;
            for (int i=0;i<curDocument.no_words;i++){
                curWord=curDocument.words.get(i);
                int wid=curWord.wid;
                if (curWord.s==0){
                    continue;
                }
                pp1=count_w_1[wid][poten_z].get();
                pp1_all=count_w_all_1[poten_z].get();
                if(poten_z==z){
                    pp1-=curDocument.word_count.get(wid);
                    pp1_all-=curDocument.count_s_1;
                }
                p1+=Math.log((beta+pp1)/(num_vocab*beta+pp1_all));
            }
            if(r==0){
                pp2=count_z[poten_z][c].get();
                if(poten_z==z) pp2--;
                pp2_all=count_z_all[c].get();
                pp2_all--;
            }else{
                pp2=count_z[poten_z][c_y].get();
                if (poten_z==z) pp2--;
                pp2_all=count_z_all[c_y].get();
                pp2_all--;
            }
            p2=(alpha+pp2)/(Z*alpha+pp2_all);
            p[poten_z]=Math.pow(Math.E,p1)*p2;
        }
        int new_z=tls.sample_multi(p);
        if (new_z!=z){
            curDocument.z=new_z;
            HashMap<Integer,Integer> wc=curDocument.word_count;
            for (int wid:wc.keySet()){
                count_w_1[wid][z].addAndGet(-wc.get(wid));
                count_w_1[wid][new_z].addAndGet(wc.get(wid));
                count_w_all_1[z].addAndGet(-wc.get(wid));
                count_w_all_1[new_z].addAndGet(wc.get(wid));
            }
            if(r==0){
                count_z[z][c].decrementAndGet();
                count_z[new_z][c].incrementAndGet();
            }else{
                count_z[z][c_y].decrementAndGet();
                count_z[new_z][c_y].incrementAndGet();
            }
        }
    }
    
    
    public void printParam(){
        try{
            System.out.println("num_users: "+num_users);
            System.out.println("num_vocab: "+num_vocab);
            System.out.println("K: "+K);
            System.out.println("Z: "+Z);
            System.out.println("n_itr: "+n_itr);
            System.out.println("burn_in: "+burn_in);
            System.out.println("sam_interval: "+sam_interval);
            System.out.println("save_dir: "+save_dir);
            System.out.println("------------------------------------");
        }catch(Exception e){
            e.printStackTrace();
        }
    }
    public void initialize(){
        User curUser;
        Document curDocument;
        Word curWord;
        Link curLink;
        //prior parameters
        beta=1.0/num_vocab;
        alpha=1.0/Z;
        zeta=1.0/K;
        eta=1.0/num_users;
        xi=1.0/num_users;
        sigma=1.0/K;
        //u=tls.random.nextDouble();//probability of word generated from background topic
        rho=tls.random.nextDouble();
	u=0.9;
        //initialize counts
        count_epsilon_1=new int[num_users];
        //count_epsilon_1=new AtomicInteger[num_users];
        //for (int i=0;i<count_epsilon_1.length;i++){
        //    count_epsilon_1[i]=new AtomicInteger(0);
        //}
        //count_epsilon_0=new int[num_users];
        count_r_1=new int[num_users];
        //count_r_1=new AtomicInteger[num_users];
        //for (int i=0;i<count_r_1.length;i++){
        //    count_r_1[i]=new AtomicInteger(0);
        //}
        count_s_1=new AtomicInteger(0);
        count_s_0=new AtomicInteger(0);
        count_lambda_1=new AtomicInteger(0);
        count_lambda_0=new AtomicInteger(0);
        //count_celeb_all=new AtomicInteger(0);
        //count_celeb=new AtomicInteger[num_users];
        //for (int i=0;i<count_celeb.length;i++){
        //    count_celeb[i]=new AtomicInteger(0);
        //}
        count_ee=new AtomicInteger[num_users][K];
        for(int i=0;i<count_ee.length;i++){
            for (int j=0;j<count_ee[i].length;j++){
                count_ee[i][j]=new AtomicInteger(0);
            }
        }
        count_ee_all=new AtomicInteger[K];
        for (int i=0;i<count_ee_all.length;i++){
            count_ee_all[i]=new AtomicInteger(0);
        }
        count_c_star=new AtomicInteger[K][K];
        for(int i=0;i<count_c_star.length;i++){
            for (int j=0;j<count_c_star[i].length;j++){
                count_c_star[i][j]=new AtomicInteger(0);
            }
        }
        count_c_star_all=new AtomicInteger[K];
        for (int i=0;i<count_c_star_all.length;i++){
            count_c_star_all[i]=new AtomicInteger(0);
        }
        count_c_y=new AtomicInteger[K][K];
        for(int i=0;i<count_c_y.length;i++){
            for (int j=0;j<count_c_y[i].length;j++){
                count_c_y[i][j]=new AtomicInteger(0);
            }
        }
        count_c_y_all=new AtomicInteger[K];
        for (int i=0;i<count_c_y_all.length;i++){
            count_c_y_all[i]=new AtomicInteger(0);
        }
        count_w_0=new AtomicInteger[num_vocab];
        for (int i=0;i<count_w_0.length;i++){
            count_w_0[i]=new AtomicInteger(0);
        }
        count_w_1=new AtomicInteger[num_vocab][Z];
        for(int i=0;i<count_w_1.length;i++){
            for (int j=0;j<count_w_1[i].length;j++){
                count_w_1[i][j]=new AtomicInteger(0);
            }
        }
        count_w_all_0=new AtomicInteger(0);
        count_w_all_1=new AtomicInteger[Z];
        for (int i=0;i<count_w_all_1.length;i++){
            count_w_all_1[i]=new AtomicInteger(0);
        }
        count_z=new AtomicInteger[Z][K];
        for(int i=0;i<count_z.length;i++){
            for (int j=0;j<count_z[i].length;j++){
                count_z[i][j]=new AtomicInteger(0);
            }
        }
        count_z_all=new AtomicInteger[K];
        for (int i=0;i<count_z_all.length;i++){
            count_z_all[i]=new AtomicInteger(0);
        }
        //initialize c_u
        c_u=new ArrayList<HashMap<Integer,Integer>>();
        for (int i=0;i<K;i++){
            c_u.add(new HashMap<Integer,Integer>());
        }
        //initialize c,c_u,v,c_y,c_star,r,z,s
        for(int i=0;i<num_users;i++){
            curUser=users.get(i);
            int new_c=tls.choice(K);
            curUser.c=new_c;
            c_u.get(new_c).put(i, 1);
            double new_v=tls.random.nextDouble();
            curUser.v=new_v;
            for (int j=0;j<curUser.no_docs;j++){
                curDocument=curUser.docs.get(j);
                int new_c_y=tls.choice(K);
                curDocument.c_y=new_c_y;
                int new_r=tls.sample_binomial(curUser.v);
                curDocument.r=new_r;
                if (new_r==1){
                    //count_r_1[i].incrementAndGet();
                    count_r_1[i]+=1;
                    count_c_y[new_c_y][new_c].incrementAndGet();
                    count_c_y_all[new_c].incrementAndGet();
                }
                int new_z=tls.choice(Z);
                curDocument.z=new_z;
                if (new_r==0){//from users' own interest
                    count_z[new_z][new_c].incrementAndGet();
                    count_z_all[new_c].incrementAndGet();
                }else{
                    count_z[new_z][new_c_y].incrementAndGet();
                    count_z_all[new_c_y].incrementAndGet();
                }
                for (int t=0;t<curDocument.no_words;t++){
                    curWord=curDocument.words.get(t);
                    int new_s=tls.sample_binomial(u);
                    //curWord.s=new_s;
                    curDocument.setS(t, new_s);
                    if (new_s==0){
                        count_s_0.incrementAndGet();
                        count_w_0[curWord.wid].incrementAndGet();
                        count_w_all_0.incrementAndGet();
                    }else{
                        count_s_1.incrementAndGet();
                        count_w_1[curWord.wid][new_z].incrementAndGet();
                        count_w_all_1[new_z].incrementAndGet();
                    }
                }
            }
            for (int j=0;j<curUser.no_links;j++){
                curLink=curUser.links.get(j);
                int new_c_star=tls.choice(K);
                curLink.c_star=new_c_star;
                int new_epsilon=tls.sample_binomial(1-curUser.v);
                int new_lambda=tls.sample_binomial(rho);
                curLink.epsilon=new_epsilon;
                curLink.lambda = new_lambda;
                if (new_epsilon==1){
                    //count_epsilon_1[i].incrementAndGet;
                    count_epsilon_1[i]+=1;
                    if (new_lambda==0){
                        count_ee[curLink.target][new_c].incrementAndGet();
                        count_ee_all[new_c].incrementAndGet();
                        count_lambda_0.incrementAndGet();
                    }else{
                        count_lambda_1.incrementAndGet();
                        //count_celeb_all.incrementAndGet();
                        //count_celeb[curLink.target].incrementAndGet();
                    }
                }else{
                    count_c_star[new_c_star][new_c].incrementAndGet();
                    count_c_star_all[new_c].incrementAndGet();
                    count_ee[curLink.target][new_c_star].incrementAndGet();
                    count_ee_all[new_c_star].incrementAndGet();
                }
            }
        }
        //initialize M,A
        M=new double[K];
        M=init_M();
        A=new double[K][num_users];
        for (int i=0;i<K;i++){
            A[i]=get_init_A(c_u.get(i).keySet());
        }
        //initialize I,theta,phi
        I=new double[K][K];
        theta=new double[K][Z];
        phi=new double[Z+1][num_vocab];//Z topic plus the background topic
        //pi=new double[num_users];
        //initialize likelihood place holders
        total_likelihood=new HashMap<Integer,Double>();
        text_likelihood=new HashMap<Integer,Double>();
        link_likelihood=new HashMap<Integer,Double>();
        likelihood_iters=new ArrayList<Integer>();
    }
    public double[] init_M(){
        double[] new_M=new double[K];
        for (int i=0;i<K;i++){
            new_M[i]=c_u.get(i).size();
        }
        return tls.normalize(new_M);
    }
    public double[] get_init_A(Set<Integer> members){
        int l_mem=members.size();//number of members in the community
        int l_oth=num_users-l_mem;//number of users that are not in the community
        double pm=(1-leave_out)/l_mem;
        double po=leave_out/l_oth;
        double[] new_result=new double[num_users];
        for(int i=0;i<num_users;i++){
            new_result[i]=po;
        }
        for (int i:members){
            new_result[i]=pm;
        }
        return new_result;
    }
}
